/*---------------------------------------------------------------------------*/
/* SCCS Information: @(#)make_bc.c	1.1    3/7/96 */
/*---------------------------------------------------------------------------*/
/* Implements the following boundary conditions
  
   periodic in x direction. 
   slip velocity 	   on y=0,1
   V.C = 0                 on y=0,1 
 
*/   
#include "utilf.h"
#include "stypes.h"
#include "markers.h"
#include "extra.h"
#include "extra1.h"
#include "make_bc.h"
#include "make_bc_periodic.h"
#include "inout.h"


void bc_vector_bound(real2D u, real2D v, int nx, int ny)
{
  int i, j;
  /* top and bottom walls */
  for (i = 1; i <= nx; i++) {
    u[i][1] = u[i][2]; /* axisymmetric */
    v[i][2] = 0.0;     /* axisymmetric */
    u[i][ny] = u[i][ny-1];
    v[i][ny] = 0.0;
  }
  /* left and right walls */
  for (j = 1; j <= ny; j++) {
    u[2][j] = 0.0;
    v[1][j] = v[2][j];
    u[nx][j] = 0.0;
    v[nx][j] = v[nx-1][j];
  }
#ifdef CENTRAL_SYMMETRY
  i = nx/2 + 1;
  for (j = 1; j <= ny; j++)
    u[i][j] = 0.0;
#endif
}


void bc_vector_div(real2D u, real2D v, real2D ap,
	           interface in,
	           int nx, int ny)
{
  int i, j;
  real eu, ev;

  /* extrapolate the interfacial cells */
  for (i = 3; i < nx - 1; i++) { /* don't solve the first and last columns */
    for (j = 2; j < ny; j++)
      if (!INU(i,j) && BU(i,j)) {
#ifdef EXTRA_LINEAR
	extra_velocity((real)i, (real)j + 0.5, u, v, ap, in, 
		       &eu, &ev, nx, ny);
#else
	extra_velocity2((real)i, (real)j + 0.5, u, v, ap, in, 
			&eu, &ev, nx, ny);
#endif
	u[i][j] = eu;
      }
      else if (!INU(i,j))
	u[i][j] = UNDEFINED;
    for (j = 3; j < ny; j++)
      if (!INV(i,j) && BV(i,j)) {
#ifdef EXTRA_LINEAR
	extra_velocity((real)i + 0.5, (real)j, u, v, ap, in, 
		       &eu, &ev, nx, ny);
#else
	extra_velocity2((real)i + 0.5, (real)j, u, v, ap, in, 
			&eu, &ev, nx, ny);
#endif
	v[i][j] = ev;
      }
      else if (!INV(i,j))
	v[i][j] = UNDEFINED;
  }
  bc_vector_bound(u, v, nx, ny);
}


void bc_pressure_inside(real2D p, real2D ap, int nx, int ny)
{
  int i, j;
  
  for (i = 2; i < nx; i++)
    for (j = 2; j < ny; j++)
      if (!INP(i,j))
	p[i][j] = 0.0;
}


void bc_pressure(real2D p, real2D ap, real pe, 
		 real R,
		 int nx, int ny)
{
  bc_pressure_inside(p, ap, nx, ny);
  bc_scalar(p, nx, ny, NULGRAD);
}


void bc_scalar(real2D scal, int nx, int ny, char sw)
{
  int i, j;
  
  /*  div: gradient equal to zero on y=0,1 planes  */
  switch(sw) {	
  case NULGRAD: case NULGRAD2:
    for (i = 1; i <= nx; i++) {
      scal[i][1] = scal[i][2]; /* axisymmetric */
      scal[i][ny] = scal[i][ny - 1];
    }     
    for (j = 1; j <= ny; j++) {
      scal[1][j] = scal[2][j];
      scal[nx][j] = scal[nx-1][j];
    }
    break;
  case NUL:
    /*  residue equal to zero on the walls  */ 
    for (i = 1; i <= nx; i++) {
      scal[i][1] = 0.0; /* axisymmetric */
      scal[i][ny] = 0.0;
    }
    for (j = 1; j <= ny; j++)
      scal[1][j] = scal[nx][j] = 0.0;
    break;
  default:
    printf("illegal switch value for a scalar field\n");
    exit(1);
  }
}


int bc_xcoord(int i, int nx)
{
  /* mirror symmetric boundary conditions */
  if (i <= 1)
    return 3 - i;
  if (i >= nx)
    return 2*nx - i - 1;
  return i;
}


int bc_ycoord(int j, int ny)
{
  /* mirror symmetric boundary conditions */
  if (j <= 1)
    return 3 - j;
  if (j >= ny)
    return 2*ny - j - 1;
  return j;
}











