
/* SCCS INFORMATION: filename - geunits.c, SCCS delta - 1.2, date - 7/22/94 */

/* package for units conversion */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <gedefine.h>
#include <geascii.h>
#include <geerror.h>

/*  NSIZ the maximum number of characters of one units item   /
/   NITEMS the number of different units in each units string /
/   and NDIM the number of dimensions                        */

#define NSIZ    10
#define NITEMS  20
#define NDIM     8

static struct table_elem {  /* units conversion table */
  char   name[NSIZ];
  double factor;
  int    dimen[NDIM];
} *table;

static struct prefix_elem { /* prefix table */
  char   name;
  double factor;
} *prefix;

static struct one_item {    /* structured list representing conversion str */
  double v;
  int    dimen[NDIM];
  char   oper;
} items[NITEMS];

static int Sntab, Snpre;      /* sizes of units and prefix tables            */
static size_t Sste, Sspe;     /* sizes of structures table_elem, prefix_elem */
static int Stab_comp(const struct table_elem *, const struct table_elem *);
static int Spre_comp(const struct prefix_elem *, const struct prefix_elem *);

/* setup units conversion tables */
void GEsetup_units(char *filename)
{
  FILE    *file;
  int     j;
  struct table_elem  *p;
  struct prefix_elem *q;

  Sste = sizeof(struct table_elem);
  Sspe = sizeof(struct prefix_elem);

  if ( (file = fopen(filename, "r")) == NULL )
    GEstandard_error("Units conversion error: failed to open units file.");

  Sntab   = GEffind_string(file,"::END CT::",TRUE) - 
    GEffind_string(file,"::CONVERSION TABLE::",TRUE) - 1;
  Snpre   = GEffind_string(file,"::END PT::",TRUE) - 
    GEffind_string(file,"::PREFIX TABLE::",TRUE) - 1;
  table  = (struct table_elem *) calloc(Sntab, Sste);
  prefix = (struct prefix_elem *) calloc(Snpre, Sspe);

  if (Sntab <= 0 OR Snpre <=0)
    GEstandard_error("Units conversion error: failed to read tables correctly.");

  GEffind_string(file,"::CONVERSION TABLE::",TRUE);  /* read the table */
  for (p=table; p<table+Sntab; p++) {
    fscanf(file, "%s %lf", &(p->name), &(p->factor));
    for (j=0; j<NDIM; j++)
      fscanf(file, "%d", &(p->dimen[j]));
    while( fgetc(file) != '\n' ) /* skip remainder of line */
      ;
  }

  GEffind_string(file,"::PREFIX TABLE::",TRUE); /* read the prefixes */
  for (q=prefix; q<prefix+Snpre; q++) {
    fscanf(file, "%s %lf", &(q->name), &(q->factor));
    while( fgetc(file) != '\n' ) /* skip remainder of line */
      ;
  }

  fclose(file);

  qsort(table,  Sntab, Sste, (int (*)(const void *,const void *))Stab_comp);
  qsort(prefix, Snpre, Sspe, (int (*)(const void *,const void *))Spre_comp);
}


/* conversion function */
double GEconvert(double v, char *from, char *to)
{
  int    i, fdimen[NDIM], tdimen[NDIM];
  double fval, tval;
  static void toSI(char *, double *, int *);
  static double fromdegK(char *, double);
  static double todegK(char *, double);

  if (!strcmp(to, "~K") OR !strcmp(to, "~C") OR
      !strcmp(to, "~R") OR !strcmp(to, "~F")) {
    fval = tval = 1.0;
    v = fromdegK(to, todegK(from, v));
  }
  else {  /* conversions for all but single temperatures */
    toSI(from, &fval, fdimen);
    toSI(to,   &tval, tdimen);    
    for (i=0; i<NDIM; i++)
      if (fdimen[i] != tdimen[i]) {
	fprintf(stderr, "\n From %s to %s\n", from, to);
	GEstandard_error("Units conversion error: inconsistent units.");
      }
  }

  return (v*fval/tval);
}

/* find conversion factor to convert a value to SI units and find associated dimensions */
static void toSI(char *s, double *v, int *dimen)
{
  int    dividing, i, n;
  struct one_item *p;
  static char *get_item(char *, struct one_item *);

  *v = 1.0;
  for (i=0; i<NDIM; i++)
    dimen[i] = 0;

  for (n=0; (s=get_item(s, items+n)) != NULL; n++)
    ;
  n++;

  for (p=items; p<items+n; p++)  /* deal with raising to a power */
    if (p->oper == '^') {
      p->oper = '*';
      p->v    = powr(p->v, (p+1)->v);
      for (i=0; i<NDIM; i++) {
	p->dimen[i] *= (int) (p+1)->v + 0.001; /* add a bit to stop round downs */
	(p+1)->dimen[i] = 0;                   /* should be zero anyway         */
      }
      (p+1)->v = 1.0;
    }

  dividing = 0;
  for (p=items; p<items+n; p++) { /* deal with mults and divisons */
    if (dividing) {
      *v /= p->v;
      for (i=0; i<NDIM; i++)
	dimen[i] -= p->dimen[i];
    }
    else {
      *v *= p->v;
      for (i=0; i<NDIM; i++)
	dimen[i] += p->dimen[i];
    }
    dividing = dividing OR (p->oper == '/');
  }
}

/* special conversions for temperature */
static double fromdegK(char *s, double v)
{
  if (!strcmp(s, "~K"))
    { /* do nothing */ }
  else if (!strcmp(s, "~C"))
    v -= 273.15;
  else if (!strcmp(s, "~R"))
    v *= 1.8;
  else if (!strcmp(s, "~F"))
    v = (v - 273.15) * 1.8 + 32.0;
  else
    GEstandard_error
      ("Units conversion error: expecting single temperature conversion");

  return (v);
}

static double todegK(char *s, double v)
{
  if (!strcmp(s, "~K"))
    { /* do nothing */ }
  else if (!strcmp(s, "~C"))
    v += 273.15;
  else if (!strcmp(s, "~R"))
    v /= 1.8;
  else if (!strcmp(s, "~F"))
    v = (v-32.0)/1.8 + 273.15;
  else
    GEstandard_error
      ("Units conversion error: expecting single temperature conversion");

  return (v);
}


/* From the string s grab one more units string and operator - convert the units */
static char *get_item(char *s, struct one_item *t)
{
  int  n;
  STR name;
  static void lookup(char *, double *, int *);

  t->v = 0.0;

  for(n=0;n<NDIM;++n)
    t->dimen[n] = 0;

  t->oper = '\0';

  while (isspace(*s))
    s++;
  for (n=0; isalnum(*s) OR (*s == '~'); n++)
    name[n] = *s++;
  name[n] = '\0';
  lookup(name, &(t->v), t->dimen);

  while (isspace(*s))
    s++;
  t->oper = *s;

  if (*s != '\0')
    return(++s);
  else
    return(NULL);
}


/* lookup a name in the units conversion tables */
static void lookup(char *name, double *v, int *dimen)
{
  struct table_elem  *t;
  struct prefix_elem *p;
  int    i;
  static int isint(char *);

  if (isint(name)) {
    *v = atoi(name);
    for (i=0; i<NDIM; i++)
      dimen[i] = 0;
  }
  else if ( (t= (struct table_elem *)
                bsearch(name, table, Sntab, Sste,
			(int (*)(const void *,const void *))Stab_comp)) !=NULL ){
    *v = t->factor;
    for (i=0; i<NDIM; i++)
      dimen[i] = t->dimen[i];
  }
  else if ( (p= (struct prefix_elem *)
                bsearch(name, prefix, Snpre, Sspe,
			(int (*)(const void *,const void *))Spre_comp)) !=NULL AND
	    (t= (struct table_elem *)
                bsearch(name+1, table, Sntab, Sste,
			(int (*)(const void *,const void *))Stab_comp)) !=NULL ){
    *v = p->factor * t->factor;
    for (i=0; i<NDIM; i++)
      dimen[i] = t->dimen[i];
  }
  else
    GEstandard_error
      (strcat("Units conversion error: failed to find units string ",name));
}


/* isint returns true if the string represnts and integer */
static int isint(char *s)
{
  while (*s != '\0')
    if (!isdigit(*s++))
      return(0);

  return(1);
}


/* comparison function for table entries */
static int Stab_comp(const struct table_elem *a, const struct table_elem *b)
{
  return(strcmp(a->name, b->name));
}

/* comparison function for prefix entries */
static int Spre_comp(const struct prefix_elem *a, const struct prefix_elem *b)
{
  return(a->name - b->name);
}



/* utility for outputting conversion tables to a file */
void GEprint_tables(char *filename)
{
  int    i, j;
  FILE   *ofile;

  if ( (ofile = fopen(filename, "w")) == NULL )
    GEstandard_error("Units conversion error: failed to output tables to file.");

  fprintf(ofile, "units table size  : %d\n", Sntab);
  fprintf(ofile, "prefix table size : %d\n", Snpre);

  fprintf(ofile, "\nUNITS TABLE\n");
  for (i=0; i<Sntab; i++) {
    fprintf(ofile, "%10s %20.12e", table[i].name, table[i].factor);
    for (j=0; j<NDIM; j++)
      fprintf(ofile, " %2d", table[i].dimen[j]);
    fprintf(ofile, "\n");
  }

  fprintf(ofile, "\nPREFIX TABLE\n");
  for (i=0; i<Snpre; i++)
    fprintf(ofile, "%c %12.4e\n", prefix[i].name, prefix[i].factor);
  fclose(ofile);
}
